'' System Version 2.003
'' OS_dspText Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/04 - Corrected windowChar, windowColor, screenChar, screenColor
''   2006/10/04 - Corrected Bsp routine for TV and lores VGA
''   2006/10/26 - Added constants for display type
''   2006/11/10 - Added routines to get color palette & foreground/background
''   2006/11/12 - Modified for use with the VGA 1024x768 Tiled Driver
''   2006/12/09 - Corrected a variety of errors related to Tiled Driver
''   2006/12/10 - Made compatible with both TV and 1024x768 Tiled VGA driver
''   2006/12/15 - Changed names of methods to return color values
''   2006/12/16 - Corrected cleareol

'' This is the windowed text driver for the Propeller OS.  It may be
'' used with either the TV driver or the 1024x768 Tiled VGA driver.
'' It is based on tv_wtext.spin version 1.2 by Phil Pilgrim which
'' is in turn based on tv_text.spin which is  2006 by Parallax, Inc.
'' Modifications  2006 by Michael Green.  Permission is given to
'' include this or portions of this program in any program for the
'' Parallax Propeller processor as long as proper credit is given.

CON
  dsp_col       =   0     'current screen column
  dsp_row       =   1     'current screen row
  dsp_color     =   2     'current color
  dsp_states    =   3     'current state flags ("lazy" eol, insert)
  dsp_wleft     =   4     'current window left column
  dsp_wtop      =   5     'current window top row
  dsp_wcols     =   6     'current window number of columns
  dsp_wrows     =   7     'current window number of rows
  dsp_window    =   8     'currently selected window
  dsp_flag      =   9     'true if storing parameters
  dsp_pcnt      =  10     'number of parameter bytes so far
  dsp_pptr      =  11     'current pointer to parameter buffer
  dsp_cog       =  12     'cog number and bitmap for vga driver cogs
  dsp_params    =  18     'parameter buffer for screen commands (6 x 1)
  dsp_mode      =  24     'display mode (1 - 16x32 pixels, 2 - 16x16 pixels)
  dsp_device    =  25     'device type (0 - TV, 1 - VGA)
  dsp_maxCols   =  26     'maximum number of columns
  dsp_maxRows   =  27     'maximum number of rows
  dsp_windows   =  28     'dsp_col through dsp_wrows for each window (8 per window)
  dspx          = 156     'end of windows information, start of display specific info

  dsp_sync      = dspx+ 0 'set to -1 at the beginning of vertical sync time
  dsp_vgaSize   = dspx+ 4

  dsp_status    = dspx+ 0 '0/1/2 = off/invisible/visible              read-only  (14 longs)
  dsp_enable    = dspx+ 4 '0/non-0 = off/on                           write-only
  dsp_pins      = dspx+ 8 '%pppmmmm = pin group, pin group mode       write-only
  dsp_tvMode    = dspx+12 '%tccip = tile,chroma,interlace,ntsc/pal    write-only
  dsp_screen    = dspx+16 'pointer to screen (words)                  write-only
  dsp_colors    = dspx+20 'pointer to colors (longs)                  write-only 
  dsp_ht        = dspx+24 'horizontal tiles                           write-only
  dsp_vt        = dspx+28 'vertical tiles                             write-only
  dsp_hx        = dspx+32 'horizontal tile expansion                  write-only 
  dsp_vx        = dspx+36 'vertical tile expansion                    write-only 
  dsp_ho        = dspx+40 'horizontal offset                          write-only  
  dsp_vo        = dspx+44 'vertical offset                            write-only
  dsp_broadcast = dspx+48 'broadcast frequency (Hz)                   write-only 
  dsp_auralcog  = dspx+52 'aural fm cog                               write-only
  dsp_tvSize    = dspx+56

  dsp_LazyEol   = %00000001                     ' dsp_status: at right end of window
  dsp_Insert    = %00000010                     ' dsp_status: insert mode (for editor)
  
'' Display control codes

  ClrWin      = $00                            ' clear window  
  Home        = $01                            ' home in window
  MoveXY      = $02                            ' move to X,Y in window (X and Y follow)
  CursLt      = $03                            ' cursor left
  CursRt      = $04                            ' cursor right
  CursUp      = $05                            ' cursor up
  CursDn      = $06                            ' cursor down
  Color       = $07                            ' select color C (0-63) (C follows)
  Bsp         = $08                            ' backspace
  Tab         = $09                            ' tab (8 spaces per)
  Lf          = $0A                            ' linefeed
  ClrEol      = $0B                            ' clear to end of line
  ClrEow      = $0C                            ' clear to end of active window
  Cr          = $0D                            ' carriage return
  MoveX       = $0E                            ' move to X in window (X follows)
  MoveY       = $0F                            ' move to Y in window (Y follows)
  SetWin      = $10                            ' set bounds of window (1-15,x,y,cols,rows)
                                               ' If cols or rows supplied as zero, use max. available
  UseWin      = $11                            ' change to selected window (0-15) (W follows)
  SetColor    = $12                            ' change color (0-63) to FG, BG (C, FG, BG follow)
                                               ' This actually changes a pair of colors for text use
                                               '  to FG,BG,FG,BG (even) and FG,FG,BG,BG (odd)
  SetColors   = $13                            ' change color (0-63) to FG3,FG2,FG1,BG (which follow)
  Insert      = $14                            ' insert (move line to right one column)
  ScrollUp    = $15                            ' scroll up the current window
  ScrollDn    = $16                            ' scroll down the current window
  WinColor    = $17                            ' change all colors in active window (C follows)
  EscChr      = $1F                            ' (ESC) print next character C as-is (C follows)
  Delete      = $7F                            ' delete current character
  Alt00       = $FF                            ' changed to $00 by out (for use in string())
  ClrScr      = Alt00                          ' equivalent to ClrWin (for use in string())

  absX        = %0000                          ' X value is relative to screen coordinates             
  relX        = %0001                          ' X value is relative to current values
  winX        = %0010                          ' X value is relative to window coordinates
  absY        = %0000                          ' Y value is relative to screen coordinates
  relY        = %0100                          ' Y value is relative to current values
  winY        = %1000                          ' Y value is relative to window coordinates

  NoChgColor  = 64                             ' color # to indicate no change to current color
  UseWinColor = 65                             ' color # to cause use of current window color

OBJ
  ldr : "OS_loader"                            ' SPIN program loader and support routines

CON
  colorMask   = ldr#colors - 1                 ' Mask for a color number
  winMask     = ldr#windows - 1                ' Mask for a window number

PUB stop | i, p, c
'' Stop TV or VGA driver - frees COGs
  p := long[ldr#videoPtr]
  repeat i from 0 to 2
    c := word[p+dsp_cog][i]
    if c
      cogstop(c >> 8)
      ldr.unregisterCOG(c & $FF)
      word[p+dsp_cog][i] := 0

PUB out(c) | i, j, k, p, f, x, y, t, m, z, cols, rows
'' Output a character.  This routine decodes the character stream, interpreting control codes.
'' These mostly follow the codes used with other text drivers with the exception of color since
'' the availability and details of color assignment vary so much from driver to driver.

  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  rows := byte[p+dsp_maxRows]
  if c == $FF
    c~
  if f := byte[p+dsp_flag]
    byte[p+dsp_params][byte[p+dsp_pptr]++] := c
    if byte[p+dsp_pptr] == byte[p+dsp_pcnt]
      case f
        MoveXY:
          updateCursor(byte[p+dsp_params][0],c,winX|winY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        SetWin:
          i := (byte[p+dsp_params][0] & winMask) << 3
          if i > 0 ' only windows 1 to ldr#windows-1 can be changed
            x := byte[p+dsp_params][1] <# (cols - 1) ' upper left col
            y := byte[p+dsp_params][2] <# (rows - 1) ' upper left row
            if (j := byte[p+dsp_params][3] <# (cols - x)) == 0
              j := cols - x ' width.  If a zero, supply maximum width
            if (k := c <# (rows - y)) == 0
              k := rows - y ' height.  If a zero, supply maximum height
            t := byte[p+dsp_color] ' use current color for new window
            long[p+dsp_windows+i][0] := 0 << 24 | t << 16 | y << 8 | x
            long[p+dsp_windows+i][1] := k << 24 | j << 16 | y << 8 | x
        UseWin:
          i := byte[p+dsp_window] << 3
          bytemove(p+dsp_windows+i, p+dsp_col, 4) ' save old values
          i := (byte[p+dsp_window] := c & winMask) << 3
          bytemove(p+dsp_col, p+dsp_windows+i, 8) ' copy new values
        SetColor:
          i := byte[p+dsp_params][0] & colorMask & !1 ' color number (even)
          j := byte[p+dsp_params][1]                  ' foreground color
          t := long[ldr#colorPtr]                     ' background color == c
          long[t][i]   := j << 24 | c << 16 | j << 8 | c ' set even and odd colors
          long[t][i+1] := j << 24 | j << 16 | c << 8 | c '  to F.B.F.B / F.F.B.B
        SetColors:
          i := byte[p+dsp_params][0] & colorMask     ' color number
          j := byte[p+dsp_params][1] << 24           ' foreground color 3
          j |= byte[p+dsp_params][2] << 16           ' foreground color 2
          j |= byte[p+dsp_params][3] << 8            ' foreground color 1
          t := long[ldr#colorPtr]                    ' background color == c
          long[t][i] := j | c
        WinColor:
          x := byte[p+dsp_wleft]
          y := byte[p+dsp_wtop]
          t := long[ldr#screenPtr]
          c := c & colorMask                         ' use specified color for window
          repeat i from y to y + byte[p+dsp_wrows] - 1
            repeat j from x to x + byte[p+dsp_wcols] - 1
              setScreenChar(j,i,-1,c)
          byte[p+dsp_color] := c
        MoveX:
          updateCursor(c,0,winX|relY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        MoveY:
          updateCursor(0,c,relX|winY,false)
        EscChr:
          print(c)
        Color:
          byte[p+dsp_color] := c & colorMask        ' use color for subsequent characters
      byte[p+dsp_flag]~
  else
    case c & $FF
      ClrWin:
        y := byte[p+dsp_wtop]
        t := byte[p+dsp_wrows]
        repeat i from y to y + t - 1
          fillwith(byte[p+dsp_wleft],i,byte[p+dsp_wcols]," ")
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      Home:     
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      MoveXY:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 2
      CursLt:
        updateCursor(-1,0,relX|relY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      CursRt:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          updateCursor(+1,0,relX|relY,true)
      CursUp:
        updateCursor(0,-1,relX|relY,false)
      CursDn:
        updateCursor(0,+1,relX|relY,false)
      SetWin, SetColors:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 5
      Bsp:
        x := byte[p+dsp_col] - byte[p+dsp_wleft]
        y := byte[p+dsp_row] - byte[p+dsp_wtop]
        if x | y
          if x
            x--
          else
            x := byte[p+dsp_wcols] - 1
            y--
          updateCursor(x,y,winX|winY,false)
          setScreenChar(byte[p+dsp_col],byte[p+dsp_row]," ",UseWinColor)
      Tab:
        repeat
          print(" ")
        while (byte[p+dsp_col] - byte[p+dsp_wleft]) & 7
      Lf:
        if (byte[p+dsp_row] - byte[p+dsp_wtop]) == byte[p+dsp_wrows] - 1
          x := byte[p+dsp_col]
          y := byte[p+dsp_row]
          newline
          updateCursor(x,y,absX|absY,false)
        else
          updateCursor(0,+1,relX|relY,false)
      ClrEol:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
      ClrEow:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
        y := byte[p+dsp_row]
        t := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
        repeat i from y + 1 to t
          fillwith(byte[p+dsp_wleft],i,byte[p+dsp_wcols]," ")
      SetColor:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 3
      Cr:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
        newline
      MoveX, MoveY, UseWin, EscChr, WinColor, Color:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 1
      Insert:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          t := long[ldr#screenPtr]
          m := byte[p+dsp_mode]
          i := byte[p+dsp_row] * m * cols
          j := byte[p+dsp_col]
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          if j < k
            repeat z from 0 to k-j-1
              word[t][i+k-z] := word[t][i+k-z-1]
            if m == 2
              repeat z from 0 to k-j-1
                word[t][i+cols+k-z] := word[t][i+cols+k-z-1]
          setScreenChar(j,byte[p+dsp_row]," ",UseWinColor)
      Delete:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          t := long[ldr#screenPtr]
          m := byte[p+dsp_mode]
          i := byte[p+dsp_row] * m * cols
          j := byte[p+dsp_col]
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          if j < k
            wordmove(t+(i+j)*2,t+(i+j+1)*2,k-j)
            if m == 2
              wordmove(t+(i+cols+j)*2,t+(i+cols+j+1)*2,k-j)
          setScreenChar(k,byte[p+dsp_row]," ",UseWinColor)
      ScrollUp:
        scrollUpWin
      ScrollDn:
        scrollDownWin
      other:
        print(c)

PUB str(stringptr)
'' Print a zero-terminated string
  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i
'' Print a decimal number
  if value < 0
    -value
    out("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB decFixed(value,width) | i, p, f, a, b, c
'' Print a decimal number of fixed width
  p := @a
  f~
  if value < 0
    -value
    byte[p++] := "-"
  i := 1_000_000_000
  repeat 10
    if value => i
      byte[p++] := value / i + "0"
      value //= i
      f~~
    elseif f or i == 1
      byte[p++] := "0"
    i /= 10
  byte[p]~
  i := strsize(@a)
  if i < width
    repeat width - i
      out(" ")
  str(@a)
  
PUB hex(value, digits)
'' Print a hexadecimal number
  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number
  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB screenChar(x, y) | p, s, c, m, cols, rows
'' Return the character at absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  rows := byte[p+dsp_maxRows]
  m := byte[p+dsp_mode]
  if x => cols or y => rows
    return " "
  s := long[ldr#screenPtr]
  c := word[s][y * m * cols + x]
  if ldr#TV
    return (c & $FE) | ((c >> 10) & 1)
  else
    return (c >> 6 & $FE) | (c & 1)

PUB screenColor(x, y) | p, s, m, cols, rows
'' Return the color of the character at absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  rows := byte[p+dsp_maxRows]
  m := byte[p+dsp_mode]
  if x => cols or y => rows
    return NoChgColor
  s := long[ldr#screenPtr]
  if ldr#TV
    return word[s][y * m * cols + x] >> 10 & colorMask & !1
  else
    return word[s][y * m * cols + x]  & colorMask & !1

PUB setScreenChar(x, y, c, z) | p, s, i, k, t, m, f, cols, rows
'' Set the character at absolute screen coordinates x,y to c with color z
'' If the character value is negative, only the color is changed.
'' If z == NoChgColor, the color is not changed.
'' If z == UseWinColor, the current window color is used.
  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  rows := byte[p+dsp_maxRows]
  m := byte[p+dsp_mode]
  if x => cols or y => rows
    return
  s := long[ldr#screenPtr]
  k := y * m * cols + x
  t := word[s][k]
  case z
    NoChgColor:
      if ldr#TV
        z := t >> 10 & !1 & colorMask
      else
        z := t & !1 & colorMask
    UseWinColor:
      z := byte[p+dsp_color]
    other:
      z := z & !1 & colorMask
  if ldr#TV
    if c < 0
      i := (t & $07FF) + (z & !1) << 10
    else
      i := (z & !1 + c & 1) << 10 + c & $FE
    i |= $0200
    f := $0001
  else
    if c < 0
      i := (t & $3F81) + z & !1
    else
      i := (c & $FE) << 6 + z & !1 + c & 1
    i |= $8000
    f := $0040
  word[s][k] := i
  if m == 2
    word[s][k+cols] := i | f

PUB windowChar(x, y) | p
'' Return the character at window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return " "
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return screenChar(x,y)

PUB windowColor(x, y) | p
'' Return the color of the character at window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return NoChgColor
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return screenColor(x,y)

PUB setWindowChar(x, y, c, z) | p
'' Set the character at window relative coordinates x,y to c with color z
'' If z == NoChgColor, the color is not changed.
'' If z == UseWinColor, the current window color is used.
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return " "
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  setScreenChar(x,y,c,z)

PUB windowAtEnd(x, y) | p
'' Return true if window relative coordinates are past boundary
  p := long[ldr#videoPtr]
  return x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]

PUB windowCol | p
'' Return the current window relative x coordinate (column)
  p := long[ldr#videoPtr]
  return byte[p+dsp_col] - byte[p+dsp_wleft]

PUB windowRow | p
'' Return the current window relative y coordinate (row)
  p := long[ldr#videoPtr]
  return byte[p+dsp_row] - byte[p+dsp_wtop]

PUB windowCols
'' Return the current window width (in columns)
  return byte[long[ldr#videoPtr]+dsp_wcols]

PUB windowRows
'' Return the current window height (in text rows)
  return byte[long[ldr#videoPtr]+dsp_wrows]

PUB out2(a, b)
'' Output a 2 character sequence
  out(a)
  out(b)

PUB out3(a, b, c)
'' Output a 3 character sequence
  out(a)
  out(b)
  out(c)

PUB out4(a, b, c, d)
'' Output a 4 character sequence
  out(a)
  out(b)
  out(c)
  out(d)

PUB out5(a, b, c, d, e)
'' Output a 5 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)

PUB out6(a, b, c, d, e, f)
'' Output a 6 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)
  out(f)

PRI print(c) | p
  p := long[ldr#videoPtr]
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    newline
  setScreenChar(byte[p+dsp_col],byte[p+dsp_row],c,UseWinColor)
  updateCursor(+1,0,relX|relY,true)

PRI updateCursor(x,y,mode,chkEol) | p, ax, ay, wx, wy, t
  p := long[ldr#videoPtr]
  ax := byte[p+dsp_wleft]
  ay := byte[p+dsp_wtop]
  wx := ax + byte[p+dsp_wcols] - 1
  wy := ay + byte[p+dsp_wrows] - 1
  case mode & %0011
    relX:
      x += byte[p+dsp_col]
      x &= x & $80000000 == 0
    winX:
      x += ax
  case (mode & %1100) >> 2
    relY >> 2:
      y += byte[p+dsp_row]
      y &= y & $80000000 == 0
    winY >> 2:
      y += ay
  if chkEol
    if x > wx or y > wy
      byte[p+dsp_states] |= dsp_LazyEol
    else
      byte[p+dsp_states] &= !dsp_LazyEol
  byte[p+dsp_col] := x := x #> ax <# wx
  byte[p+dsp_row] := y := y #> ay <# wy

PRI newline | p
  p := long[ldr#videoPtr]
  updateCursor(0,+1,winX|relY,true)
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    scrollUpWin
  byte[p+dsp_states] &= !dsp_LazyEol

PRI scrollUpWin | i, k, p, e
  p := long[ldr#videoPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from byte[p+dsp_wtop] to e - 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i+1)
    k := byte[p+dsp_col]
    updateCursor(0,e,winX|absY,false)
    cleartoeol
    updateCursor(k,0,absX|relY,false)

PRI scrollDownWin | i, k, p, e
  p := long[ldr#videoPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from e to byte[p+dsp_wtop] + 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i-1)
    k := byte[p+dsp_col]
    updateCursor(0,0,winX|winY,false)
    cleartoeol
    updateCursor(k,0,absX|winY,false)

PRI cleartoeol | p, x, w
  p := long[ldr#videoPtr]
  x := byte[p+dsp_col]
  w := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
  fillwith(x,byte[p+dsp_row],w-x+1," ")
       
PRI fillwith(x, y, w, c) | p, i, s, m, f, z, cols
  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  m := byte[p+dsp_mode]
  s := long[ldr#screenPtr]
  z := byte[p+dsp_color]
  if ldr#TV
    c := $0200 + (z & !1 + c & 1) << 10 + c & $FE
    f := $0001
  else
    c := $8000 + (c & $FE) << 6 + z & !1 + c & 1
    f := $0040
  y := y * m * cols
  wordfill(s+(y+x)*2,c,w)
  if m == 2
    wordfill(s+(y+x+cols)*2,c|f,w)

PRI moveline(x, y, w, f) | p, i, s, m, cols
  p := long[ldr#videoPtr]
  cols := byte[p+dsp_maxCols]
  m := byte[p+dsp_mode]
  s := long[ldr#screenPtr]
  y := y * m * cols
  f := f * m * cols
  wordmove(s+(y+x)*2,s+(f+x)*2,w)
  if m == 2
    wordmove(s+(y+x+cols)*2,s+(f+x+cols)*2,w)

PUB colorPalette                               '' Return current color palette #
  return byte[long[ldr#videoPtr]+dsp_color]

PUB pixel0Color(c)                             '' Return pixel color 0 for c
  return long[long[ldr#colorPtr]][c] & $FF

PUB pixel1Color(c)                             '' Return pixel color 1 for c
  return (long[long[ldr#colorPtr]][c] >> 8) & $FF

PUB pixel2Color(c)                             '' Return pixel color 2 for c
  return (long[long[ldr#colorPtr]][c] >> 16) & $FF

PUB pixel3Color(c)                             '' Return pixel color 3 for c
  return (long[long[ldr#colorPtr]][c] >> 24) & $FF

PUB foregroundColor(c)                         '' Return foreground color for c
  return (long[long[ldr#colorPtr]][c] >> ((c & 1) << 4)) & $FF

PUB backgroundColor(c)                         '' Return background color for c
  return long[long[ldr#colorPtr]][c] & $FF

PUB maxCols                                    '' Return display width (in columns)
  return byte[long[ldr#videoPtr]+dsp_maxCols]

PUB maxRows                                    '' Return the display height (in text rows)
  return byte[long[ldr#videoPtr]+dsp_maxRows]

PUB device                                     '' Return the display device type code
  return byte[long[ldr#videoPtr]+dsp_device]   '  0 - TV, 1 - VGA