'' System Version 2.003
'' OS_initialization Changes:
''
''   2006/09/24 - Moved from the main program to a separate object.  Added a parameter
''                for the initial display message.
''   2006/11/02 - Modified for use with the Hydra system
''   2006/11/03 - Now makes sure boot EEPROM is called "**BOOT**" internally
''   2006/11/04 - Call ldrInit.computeTimes before calling ldr.start
''   2006/11/06 - Initialize relocated on-board boot EEPROM area if not already done
''   2006/11/08 - Changed fatal error message editing
''   2006/11/10 - Now does no initialization if already done once
''   2006/11/13 - Keyboard and display pin groups moved to OS_loader
''   2006/11/27 - Added initialization for Vinculum USB Host
''   2006/12/08 - Cleaned up display initialization code
''   2006/12/12 - Modified to include both video drivers and select which to start
''                Modified to include both keyboard drivers and select which to start
''   2006/12/16 - Changed start and showBanner to get identification info from call

'' This is the initialization routine for the Propeller Operating System.  It gets its
'' hardware specific information from OS_loader, so this code is hardware independent.

OBJ
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  ldrInit : "OS_loaderInit"                    '  Initialization and assembly portion
  dsp     : "OS_dspText"                       ' Display text driver modified for OS use
  tvInit  : "OS_tvTextInit"                    '  Initialization and assembly portion for TV
  vgaInit : "OS_vgaTextInit"                   '  Initialization and assembly portion for VGA
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  dkyInit : "OS_demoKeyInit"                   '  Initialization and assembly portion for DemoBoard
  hkyInit : "OS_hydraKeyInit"                  '  Initialization and assembly portion for Hydra
  files   : "OS_fileSystem"                    ' File system support
'  mass    : "OS_massStorage"                   ' DOSonChip mass storage support

PUB start(name,version,release,testLevel,extra) | t, i, p
'' Operating System initialization.  Only check for keyboard and presence of boot EEPROM
'' the first time.  A marker is placed in high memory (at ldr#initMarker) to indicate
'' that initialization has already occurred.  This is erased during the boot process.
  if long[ldr#initMarker] <> ldr#uniqueMark                ' Check for a valid initialization
    repeat i from 0 to 7
      if i <> cogid                                        ' Stop all cogs except this one
        cogstop(i)
    long[ldr#memPtr]    := ldr#endFree                     ' Start at end of HUB memory
    p := ldr.allocatePRI(ldr#systemSize)                   ' Allocate system work area
    long[ldr#systemPtr] := p
    long[p+ldr#sys_cog] := |<cogid                         ' Mark current cog in use
    byte[p+ldr#sys_cogMain] := cogid + 1
    long[p+ldr#sys_random] := ldr#initMarker               ' Initialize random seed
    wordfill(p+ldr#sys_eeproms,0,ldr#maxPinPairs)          ' Clear EEPROMs bitmap table
    long[ldr#videoPtr]  := ldr.allocatePRI(ldr#videoSize)  ' Allocate work area for video
    long[ldr#screenPtr] := ldr.allocatePRI(ldr#screenSize) ' Allocate screen buffer
    long[ldr#colorPtr]  := ldr.allocatePRI(ldr#colorSize)  ' Allocate screen color buffer
    long[ldr#fontPtr]   := ldr.allocatePRI(ldr#fontSize)   ' Allocate font table
    long[ldr#cursorPtr] := ldr.allocatePRI(ldr#cursorSize) ' Allocate cursor control block
    case ldr#device
      ldr#tvDevice:                                        ' Start up the TV driver if requested
        if not tvInit.start(ldr#dspPinGroup,ldr#device,ldr#cols,ldr#rows,ldr#tiles)
          abort false
      ldr#vgaDevice:                                       ' Start up the VGA driver if requested
        if not vgaInit.start(ldr#dspPinGroup,ldr#device,ldr#cols,ldr#rows,ldr#tiles)
          abort false
    showBanner(name,version,release,testLevel,extra)       ' Display a "sign on" string
    long[ldr#keybdPtr]  := ldr.allocatePRI(ldr#keybdSize)  ' Allocate work area for keyboard
    if ldr#Parallax
      if not dkyInit.start(ldr#keyPinGroup)                ' Now start up the keyboard driver for
        abort false                                        '  the appropriate type of hardware
    if ldr#Hydra
      if not hkyInit.start(ldr#keyPinGroup)
        abort false
    ldrInit.computeTimes                                   ' Initialize I2C timing values
    if not ldr.start(ldrInit.getEntryPoint)                ' Now start up the I2C driver
      abort false
    long[ldr#userPtr] := ldr#noSuchAddr
    long[ldr#initMarker] := ldr#uniqueMark                 ' Basic initialization complete
    waitcnt(clkfreq * 2 + cnt)                             ' Wait for keyboard to initialize
    dsp.out(dsp#Cr)
    if not key.present
      abort initError(string("--> keyboard not present",dsp#Cr),0)      
    if (t := \files.updateBitmapEEPROMs(ldr#i2cBootSCL)) < 0
      return initError(string("--> can't access boot EEPROM!"),t)
    if (t & $4000) == 0                                    ' Check boot EEPROM present
      return initError(string("--> no boot EEPROM present"),0)
    if (t := \files.writeFilename(0,ldr#bootAddr,0)) < 0   ' Make sure boot EEPROM is **BOOT
      return initError(string("--> can't write boot EEPROM!"),t)
    if ldr.checkPresence(ldr#altBootAddr)                  ' See if cartridge in place
      if (t := \files.checkSums(string("**BOOT      "),ldr#altBootAddr)) < 0
        if t == true
          t := 0
        else
          return initError(string("--> error in alt boot EEPROM!"),t)
      if (t := \files.writeFilename(0,ldr#altBootAddr,t)) < 0
        return initError(string("--> can't write alt boot EEPROM!"),t)
 '   if (t := \mass.start) <> true
 '     case t
 '       mass#Timeout:
 '         return initError(string("--> mass storage Timeout"),0)
 '       mass#Invalid:
 '         return initError(string("--> mass storage Invalid"),0)
 '     return initError(string("--> mass storage CF"),0)
  return true

PRI initError(m,t)                                         '' Display error message
  dsp.str(m)                                               '   then disable OS
  if t <> 0
    dsp.str(string(dsp#Cr,"    "))
    dsp.str(-t)
  dsp.out(dsp#Cr)
  long[ldr#initMarker] := 0
  return false

PUB showBanner(name,version,release,testLevel,extra)       '' Display banner
  dsp.str(name)
  if version > 0 or release > 0 or testLevel > 0
    dsp.str(string(" Version "))
    dsp.dec(version)
    dsp.out(".")
    if release < 100
      dsp.out("0")
    if release < 10
      dsp.out("0")
    dsp.dec(release)
    if testLevel > 0
      dsp.out("a"+testLevel-1)
  if extra
    dsp.out(dsp#Cr)
    dsp.str(extra)
    