'' Modified Tiny Basic for use with Propeller Demo Board.
'' Original Copyright 2006, Radical Eye Software.

obj
   init  : "SB_initialization"         ' OS Initialization routines and definitions
   ldr   : "OS_loader"                 ' SPIN program loader and support routines
   dsp   : "OS_dspText"                ' Display text driver modified for OS use
   key   : "OS_keyboard"               ' PS/2 keyboard driver modified for OS use
   fsrw  : "fsrw"                      ' FAT File System Read/Write

con
   progsize  = 6188                    ' Space reserved for program

  _clkmode  = ldr#clockMode
  _xinfreq  = ldr#xtalClock
  OS_min    = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4
  _free     = OS_min + progsize / 4    ' Minimum for I/O + program
  _stack    = 100 * 4                  ' Roughly 4 nested expressions

   version   = 1
   release   = 4
   testLevel = 0

   spiDO     = (ldr#Hydra & 16) | (ldr#Parallax & 0)  ' SD Card SPI Data Out
   spiClk    = (ldr#Hydra & 17) | (ldr#Parallax & 1)  ' SD Card SPI Clock
   spiDI     = (ldr#Hydra & 18) | (ldr#Parallax & 2)  ' SD Card SPI Data In
   spiCS     = (ldr#Hydra & 19) | (ldr#Parallax & 3)  ' SD Card SPI Chip Select
   
   fReturn   = 13
   fLinefeed = 10
   escapeKey = $CB
      
   maxstack  = 20                     ' Maximum stack depth
   ntoks     = 53                     ' Number of tokens
   linelen   = 256                    ' Maximum input line length
   quote     = 34                     ' Double quote
   caseBit   = !32                    ' Uppercase/Lowercase bit
   
var
   long sp, tp, eop, nextlineloc, rv, curlineno, pauseTime
   long vars[26], stack[maxstack], control[2]
   long forStep[26], forLimit[26], forLoop[26]
   byte tline[linelen], inVars[26]

dat
   tok0  byte "IF", 0
   tok1  byte "THEN", 0
   tok2  byte "INPUT", 0    ' INPUT {"<prompt>";} <var> {,<var>}
   tok3  byte "PRINT", 0    ' PRINT {USING "<format>";} ...
   tok4  byte "GOTO", 0
   tok5  byte "GOSUB", 0
   tok6  byte "RETURN", 0
   tok7  byte "REM", 0
   tok8  byte "NEW", 0
   tok9  byte "LIST", 0
   tok10 byte "RUN", 0
   tok11 byte "RND", 0
   tok12 byte "OPEN", 0     ' OPEN " <file> ",<mode>
   tok13 byte "READ", 0     ' READ <var> {,<var>}
   tok14 byte "WRITE", 0    ' WRITE {USING "<format>";} ...
   tok15 byte "CLOSE", 0    ' CLOSE
   tok16 byte "DELETE", 0   ' DELETE " <file> "
   tok17 byte "RENAME", 0   ' RENAME " <file> "," <file> "
   tok18 byte "FILES", 0    ' FILES
   tok19 byte "SAVE", 0     ' SAVE { " <file> " }
   tok20 byte "LOAD", 0     ' LOAD { " <file> " }
   tok21 byte "NOT" ,0      ' NOT <logical>
   tok22 byte "AND" ,0      ' <logical> AND <logical>
   tok23 byte "OR", 0       ' <logical> OR <logical>
   tok24 byte "SHL", 0      ' <expr> SHL <expr>
   tok25 byte "SHR", 0      ' <expr> SHR <expr>
   tok26 byte "FOR", 0      ' FOR <var> = <expr> TO <expr>
   tok27 byte "TO", 0
   tok28 byte "STEP", 0     '  optional STEP <expr>
   tok29 byte "NEXT", 0     ' NEXT <var>
   tok30 byte "INA", 0      ' INA [ <expr> ]
   tok31 byte "OUTA", 0     ' OUTA [ <expr> ] = <expr>
   tok32 byte "PAUSE", 0    ' PAUSE <time ms> {,<time us>}
   tok33 byte "USING", 0    ' PRINT USING "<format>"; ...
   tok34 byte "ROL", 0      ' <expr> ROL <expr>
   tok35 byte "ROR", 0      ' <expr> ROR <expr>
   tok36 byte "SAR", 0      ' <expr> SAR <expr>
   tok37 byte "REV", 0      ' <expr> REV <expr>
   tok38 byte "BYTE", 0     ' BYTE [ <expr> ]
   tok39 byte "WORD", 0     ' WORD [ <expr> ]
   tok40 byte "LONG", 0     ' LONG [ <expr> ]
   tok41 byte "CNT", 0
   tok42 byte "PHSA", 0
   tok43 byte "PHSB", 0
   tok44 byte "FRQA", 0
   tok45 byte "FRQB", 0
   tok46 byte "CTRA", 0
   tok47 byte "CTRB", 0
   tok48 byte "DISPLAY", 0 ' DISPLAY <expr> {,<expr>}
   tok49 byte "KEYCODE", 0 ' KEYCODE
   tok50 byte "LET", 0
   tok51 byte "STOP", 0
   tok52 byte "END", 0

   toks  long @tok0, @tok1, @tok2, @tok3, @tok4, @tok5, @tok6, @tok7
         long @tok8, @tok9, @tok10, @tok11, @tok12, @tok13, @tok14, @tok15
         long @tok16, @tok17, @tok18, @tok19, @tok20, @tok21, @tok22, @tok23
         long @tok24, @tok25, @tok26, @tok27, @tok28, @tok29, @tok30, @tok31
         long @tok32, @tok33, @tok34, @tok35, @tok36, @tok37, @tok38, @tok39
         long @tok40, @tok41, @tok42, @tok43, @tok44, @tok45, @tok46, @tok47
         long @tok48, @tok49, @tok50, @tok51, @tok52

   syn   byte "Syntax Error", 0
   ln    byte "Invalid Line Number", 0
   title byte "FemtoBasic", 0

PUB main | err
'' Initialize the operating system, allocate space for the Basic program,
'' clear the program space and variables, then read a line and interpret it.
  init.start(@title,version,release,testLevel,0)
  ldr.allocate(progsize)
  clearall
  repeat
    err := \doline
    if err
      if curlineno => 0
        dsp.str(string("IN LINE "))
        dsp.dec(curlineno)
        dsp.out(" ")
      putlinet(err)
      nextlineloc := eop - 2

PRI bgetc | c, x, y, p
   repeat
      x := dsp.windowCol
      y := dsp.windowRow
      p := dsp.windowChar(x,y)
      dsp.setWindowChar(x,y,$5F,dsp#NoChgColor)    ' display underline
      c := key.getKey
      dsp.setWindowChar(x,y,p,dsp#NoChgColor)      ' Restore character
      case c
         $c8: return dsp#Bsp
         13, " ".."~":
            return c

PRI getline | i, c
   i := 0
   repeat
      c := bgetc
      if c == dsp#Bsp
         if i > 0
            dsp.out(dsp#Bsp)
            dsp.out(" ")
            dsp.out(dsp#Bsp)
            i--
      elseif c == dsp#Cr
         dsp.out(dsp#Cr)
         tline[i] := 0
         tp := @tline
         return
      elseif i < linelen-1
         dsp.out(c)
         tline[i++] := c

pri putlinet(s) | c
   repeat while c := byte[s++][0]
      if c => 128
         dsp.str(@@toks[c-128])
         dsp.out(" ")
      else
         dsp.out(c)
   dsp.out(dsp#Cr)

pri spaces | c
   repeat
      c := byte[tp]
      if c == 0 or c > " "
         return c
      tp++

pri skipspaces
   if byte[tp]
      tp++
   return spaces

pri parseliteral | r, c
   r := 0
   repeat
      c := byte[tp]
      if c < "0" or c > "9"
         return r
      r := r * 10 + c - "0"
      tp++

pri movprog(at, delta)
   if eop + delta + 2 - long[ldr#userPtr] > progsize
      abort string("NO MEMORY")
   bytemove(at+delta, at, eop-at)
   eop += delta

pri fixvar(c)
   if c => "a"
      c -= 32
   return c - "A"

pri isvar(c)
   c := fixvar(c)
   return c => 0 and c < 26

pri tokenize | tok, c, at, put, state, i, j
   at := tp
   put := tp
   state := 0
   repeat while c := byte[at]
      if c == quote
         if state == "Q"
            state := 0
         elseif state == 0
            state := "Q"
      if state == 0
         repeat i from 0 to constant(ntoks-1)
            tok := @@toks[i]
            j := 0
            repeat while byte[tok] and ((byte[tok] ^ byte[j+at]) & caseBit) == 0
               j++
               tok++
            if byte[tok] == 0 and not isvar(byte[j+at])
               byte[put++] := 128 + i
               at += j
               if i == 7
                  state := "R"
               else
                  repeat while byte[at] == " "
                     at++
               next
      byte[put++] := byte[at++]
   byte[put] := 0

pri wordat(loc)
   return (byte[loc]<<8)+byte[loc+1]

pri findline(lineno) | at
   at := long[ldr#userPtr]
   repeat while wordat(at) < lineno
      at += 3 + strsize(at+2)
   return at

pri insertline | lineno, fc, loc, locat, newlen, oldlen
   lineno := parseliteral
   if lineno < 0 or lineno => 65535
      abort @ln
   tokenize
   fc := spaces
   loc := findline(lineno)
   locat := wordat(loc)
   newlen := 3 + strsize(tp)
   if locat == lineno
      oldlen := 3 + strsize(loc+2)
      if fc == 0
         movprog(loc+oldlen, -oldlen)
      else
         movprog(loc+oldlen, newlen-oldlen)
   elseif fc
      movprog(loc, newlen)
   if fc
      byte[loc] := lineno >> 8
      byte[loc+1] := lineno
      bytemove(loc+2, tp, newlen-2)

pri clearvars
   sp := 0
   nextlineloc := long[ldr#userPtr]
   bytefill(@vars, 0, 26)
   pauseTime := 0

pri clearall
   byte[long[ldr#userPtr]][0] := 255
   byte[long[ldr#userPtr]][1] := 255
   byte[long[ldr#userPtr]][2] := 0
   eop := long[ldr#userPtr] + 2
   clearvars

pri pushstack
   if sp => constant(maxstack-1)
      abort string("RECURSION ERROR")
   stack[sp++] := nextlineloc

pri getAddress(range) | t
   if spaces <> "["
      abort @syn
   skipspaces
   result := expr
   if range
      if result < 0 or result > 31
         abort string("Invalid pin number")
      if spaces == "."
         if byte[++tp] <> "."
           abort @syn
         result <<= 8
         skipspaces
         t := expr
         if t < 0 or t > 31
            abort string("Invalid pin number")
         result |= t | $10000
   if spaces <> "]"
      abort @syn
   tp++

pri factor | tok, t
   tok := spaces
   tp++
   case tok
      "(":
         t := expr
         if spaces <> ")"
            abort @syn
         tp++
         return t
      "a".."z","A".."Z":
         return vars[fixvar(tok)]
      158: ' INA [ <expr>{..<expr>} ]
         t := getAddress(true)
         if t > $FFFF
           tok := t & $FF
           t := (t >> 8) & $FF
           dira[t..tok]~
           return ina[t..tok]
         else
           dira[t]~
           return ina[t]
      166: ' BYTE [ <expr> ]
         return byte[getAddress(false)]
      167: ' WORD [ <expr> ]
         return word[getAddress(false)]
      168: ' LONG [ <expr> ]
         return long[getAddress(false)]
      169: ' CNT
         return CNT
      170: ' PHSA
         return PHSA
      171: ' PHSB
         return PHSB
      172: ' FRQA
         return FRQA
      173: ' FRQB
         return FRQB
      174: ' CTRA
         return CTRA
      175: ' CTRB
         return CTRB
      177: ' KEYCODE
         return key.key
      139: ' RND
         return (rv? >> 1) ** (factor << 1)
      "-":
         return - factor
      "!":
         return ! factor
      "$", "%", quote, "0".."9":
         --tp
         return getAnyNumber
      other:
         abort(@syn)

pri shifts | tok, t
   t := factor
   tok := spaces
   if tok == 152 ' SHL
      tp++
      return t << factor
   elseif tok == 153 ' SHR
      tp++
      return t >> factor
   elseif tok == 162 ' ROL
      tp++
      return t <- factor
   elseif tok == 163 ' ROR
      tp++
      return t -> factor
   elseif tok == 164 ' SAR
      tp++
      return t ~> factor
   elseif tok == 165 ' REV
      tp++
      return t >< factor
   else
      return t

pri bitFactor | tok, t
   t := shifts
   repeat
      tok := spaces
      if tok == "&"
         tp++
         t &= shifts
      else
         return t

pri bitTerm | tok, t
   t := bitFactor
   repeat
      tok := spaces
      if tok == "|"
         tp++
         t |= bitFactor
      elseif tok == "^"
         tp++
         t ^= bitFactor
      else
         return t

pri term | tok, t
   t := bitTerm
   repeat
      tok := spaces
     if tok == "*"
        tp++
        t *= bitTerm
     elseif tok == "/"
        if byte[++tp] == "/"
           tp++
           t //= bitTerm
        else
           t / =bitTerm
     else
        return t

pri arithExpr | tok, t
   t := term
   repeat
      tok := spaces
      if tok == "+"
         tp++
         t += term
      elseif tok == "-"
         tp++
         t -= term
      else
         return t

pri compare | op, a, b, c
   a := arithExpr
   op := 0
   spaces
   repeat
      c := byte[tp]
      case c
         "<": op |= 1
              tp++
         ">": op |= 2
              tp++
         "=": op |= 4
              tp++
         other: quit
   case op
      0: return a
      1: return a < arithExpr
      2: return a > arithExpr 
      3: return a <> arithExpr
      4: return a == arithExpr
      5: return a => arithExpr
      6: return a =< arithExpr
      7: abort string("Invalid comparison")

pri logicNot | tok
   tok := spaces
   if tok == 149 ' NOT
      tp++
      return not compare
   return compare

pri logicAnd | t, tok
   t := logicNot
   repeat
      tok := spaces
      if tok == 150 ' AND
         tp++
         t := t and logicNot
      else
         return t

pri expr | tok, t
   t := logicAnd
   repeat
      tok := spaces
      if tok == 151 ' OR
         tp++
         t := t or logicAnd
      else
         return t

pri specialExpr
   if spaces <> "="
      abort @syn
   skipspaces
   return expr

pri scanFilename(f) | c, chars
   chars := 0
   tp++ ' skip past initial quote
   repeat while (c := byte[tp++]) <> quote
      if chars++ < 11
         byte[f++] := c
   byte[f] := 0

pri texec | ht, nt, a, b, c, d, restart, uS, thisLine, format
   uS := clkfreq / 1_000_000
   thisLine := tp - 2
   restart := 1
   repeat while restart
      restart := 0
      ht := spaces
      if ht == 0
         return
      nt := skipspaces
      if isvar(ht) and nt == "="
         tp++
         vars[fixvar(ht)] := expr
      elseif ht == ":"
         restart := 1
      elseif ht => 128
         case ht
            128: ' if
               a := expr
               if spaces <> 129
                  abort string("MISSING THEN")
               skipspaces
               if not a
                  return
               restart := 1
            130: ' input {"<prompt>";} <var> {, <var>}
               if nt == quote
                  c := byte[++tp]
                  repeat while c <> quote and c
                     dsp.out(c)
                     c := byte[++tp]
                  if c <> quote
                     abort @syn
                  if skipspaces <> ";"
                     abort @syn
                  nt := skipspaces
               if not isvar(nt)
                  abort @syn
               d := 0
               inVars[d++] := fixvar(nt)
               repeat while skipspaces == ","
                  nt := skipspaces
                  if not isvar(nt) or d == 26
                     abort @syn
                  inVars[d++] := fixvar(nt)
               getline
               tokenize
               vars[inVars[0]] := expr
               repeat a from 1 to d - 1
                  if spaces <> ","
                     abort @syn
                  skipspaces
                  vars[inVars[a]] := expr
            131: ' print
               repeat
                  nt := spaces
                  if nt == 0 or nt == ":"
                     quit
                  if nt == quote
                     tp++
                     repeat
                        c := byte[tp++]
                        if c == 0 or c == quote
                           quit
                        dsp.out(c)
                  else
                     dsp.dec(expr)
                  nt := spaces
                  if nt == ";"
                     tp++
                  elseif nt == ","
                     dsp.out(" ")
                     repeat while dsp.windowCol & 7
                        dsp.out(" ")
                     tp++
                  elseif nt == 0 or nt == ":"
                     dsp.out(dsp#Cr)
                     quit
                  else
                     abort @syn
            132, 133: ' goto. gosub
               a := expr
               if a < 0 or a => 65535
                  abort @ln
               b := findline(a)
               if wordat(b) <> a
                  abort @ln
               if ht == 133
                  pushstack
               nextlineloc := b 
            134: ' return
               if sp == 0
                  abort string("INVALID RETURN")
               nextlineloc := stack[--sp]
            136: ' new
               clearall
            137: ' list
               a := long[ldr#userPtr]
               repeat while a+2 < eop
                  dsp.dec(wordat(a))
                  dsp.out(" ")
                  putlinet(a+2)
                  a += 3 + strsize(a+2)
            138: ' run
                  clearvars
            140: ' open " <file> ", R/W/A
               if spaces <> quote
                  abort @syn
               scanFilename(@a)
               if spaces <> ","
                  abort @syn
               case skipspaces
                  "A", "a": d := "a"
                  "W", "w": d := "w"
                  "R", "r": d := "r"
                  other: abort string("Invalid open file mode")
               tp++
               if fsrw.mount(spiDO)
                  abort string("Can't mount SD card")
               if \fsrw.popen(@a,d)
                  abort string("Can't open file")
            141: ' read <var> {, <var> }
               if not isvar(nt)
                  abort @syn
               d := 0
               inVars[d++] := fixvar(nt)
               repeat while skipspaces == ","
                  if not isvar(nt) or d == 26
                     abort @syn
                  inVars[d++] := fixvar(nt)
               a := 0
               repeat
                  c := fsrw.pgetc
                  if c < 0
                     abort string("Can't read file")
                  elseif c == fReturn
                     tline[a] := 0
                     tp := @tline
                     quit
                  elseif c == fLinefeed
                     next
                  elseif a < linelen-1
                     tline[a++] := c
               tokenize
               vars[inVars[0]] := expr
               repeat a from 1 to d - 1
                  if spaces <> ","
                     abort @syn
                  skipspaces
                  vars[inVars[a]] := expr
            142: ' write ...
               d := 0 ' record column
               repeat
                  nt := spaces
                  if nt == 0 or nt == ":"
                     quit
                  if nt == quote
                     tp++
                     repeat
                        c := byte[tp++]
                        if c == 0 or c == quote
                           quit
                        fsrw.pputc(c)
                        d++
                  else
                     a := expr
                     if a < 0
                        -a
                        fsrw.pputc("-")
                     b := 1_000_000_000
                     c := false
                     repeat 10
                        if a => b
                           fsrw.pputc(a / b + "0")
                           a //= b
                           c := true
                        elseif c or b == 1
                           fsrw.pputc("0")
                        b /= 10
                  nt := spaces
                  if nt == ";"
                     tp++
                  elseif nt == ","
                     fsrw.pputc(" ")
                     d++
                     repeat while d & 7
                        fsrw.pputc(" ")
                        d++
                     tp++
                  elseif nt == 0 or nt == ":"
                     fsrw.pputc(fReturn)
                     fsrw.pputc(fLinefeed)
                     quit
                  else
                     abort @syn
            143: ' close
               if fsrw.pclose < 0
                  abort string("Error deleting file")
            144: ' delete " <file> "
               if spaces <> quote
                  abort @syn
               scanFilename(@a)
               if fsrw.mount(spiDO)
                  abort string("Can't mount SD card")
               if \fsrw.popen(@a,"d")
                  abort string("Can't delete file")
               if fsrw.pclose < 0
                  abort string("Error deleting file")
            145: ' rename " <file> "," <file> "
               if spaces <> quote
                  abort @syn
               scanFilename(@a)
               if spaces <> ","
                  abort @syn
               if skipspaces <> quote
                  abort @syn
               scanFilename(@a)
               abort string("Rename not implemented")
            146: ' files
               if fsrw.mount(spiDO)
                  abort string("Can't mount SD card")
               fsrw.opendir
               repeat while fsrw.nextfile(@a) == 0
                  dsp.str(@a)
                  dsp.out(dsp#Cr)
            147: ' save or save "<filename>"
               d := eop - long[ldr#userPtr]
               if (nt := spaces) == quote
                  scanFilename(@a)
                  if fsrw.mount(spiDO)
                     abort string("Can't mount SD card")
                  if \fsrw.popen(@a,"w")
                     abort string("Can't create file")
                  if fsrw.pwrite(@d,2) <> 2
                     abort string("Can't save program size")
                  if fsrw.pwrite(long[ldr#userPtr],d) <> d
                     abort string("Can't save program")
                  if fsrw.pclose < 0
                     abort string("Error closing file")
               else
                  if nt
                     abort @syn
                  a := ldr#i2cBootSCL << 18 + $FFF0 - progsize
                  ldr.writeEEPROM(a-2,@d,2)      ' Write program size, then
                  ldr.writeWait(a-2)             '  save the program itself
                  repeat c from 0 to d step 16
                     ldr.writeEEPROM(a+c, long[ldr#userPtr]+c, d-c <# 16)
                     ldr.writeWait(a+c)
            148: ' load or load "<filename>"
               if (nt := spaces) == quote
                  scanFilename(@a)
                  if fsrw.mount(spiDO)
                     abort string("Can't mount SD card")
                  if \fsrw.popen(@a,"r")
                     abort string("Can't open file")
                  if fsrw.pread(@d,2) <> 2
                     abort string("Can't read program size")
                  if d < 3 or d > progsize
                     abort string("Invalid program size")
                  if fsrw.pread(long[ldr#userPtr],d) <> d
                     abort string("Can't read program")
                  if fsrw.pclose < 0
                     abort string("Error closing file")
               else
                  if nt
                     abort @syn                  ' Read program from EEPROM
                  a := ldr#i2cBootSCL << 18 + $FFF0 - progsize
                  ldr.readEEPROM(a-2,@d,2)       ' First read program size
                  if d < 3 or d > progsize
                     abort string("Invalid program size")
                  ldr.readEEPROM(a,long[ldr#userPtr],d)  ' Read in actual program
               eop := long[ldr#userPtr] + d
               nextlineloc := eop - 2            ' Leave it stopped
            154: ' for <var> = <expr> to <expr> {step <expr>}
               ht := spaces
               if ht == 0
                  abort @syn
               nt := skipspaces
               if not isvar(ht) or nt <> "="
                  abort @syn
               a := fixvar(ht)
               skipspaces
               vars[a] := expr
               if spaces <> 155 ' TO             ' Save FOR limit
                  abort @syn
               skipspaces
               forLimit[a] := expr
               if spaces == 156 ' STEP           ' Save step size
                  skipspaces
                  forStep[a] := expr
               else
                  forStep[a] := 1                ' Default step is 1
               if spaces
                  abort @syn
               forLoop[a] := nextlineloc         ' Save address of line
               if forStep[a] < 0                 '  following the FOR
                  b := vars[a] => forLimit[a]
               else                              ' Initially past the limit?
                  b := vars[a] =< forLimit[a]
               if not b                          ' Search for matching NEXT 
                  repeat while nextlineloc < eop-2
                     curlineno := wordat(nextlineloc)
                     tp := nextlineloc + 2
                     nextlineloc := tp + strsize(tp) + 1
                     if spaces == 157            ' NEXT <var>
                        nt := skipspaces         ' Variable has to agree
                        if not isvar(nt)
                           abort @syn
                        if fixvar(nt) == a       ' If match, continue after
                           quit                  '  the matching NEXT
            157: ' next <var>
               nt := spaces
               if not isvar(nt)
                  abort @syn
               a := fixvar(nt)
               vars[a] += forStep[a]             ' Increment or decrement the
               if forStep[a] < 0                 '  FOR variable and check for
                  b := vars[a] => forLimit[a]
               else                              '  the limit value
                  b := vars[a] =< forLimit[a]
               if b                              ' If continuing loop, go to
                  nextlineloc := forLoop[a]      '  statement after FOR
            159: ' outa [ <expr>{..<expr>} ] = <expr>
               a := getAddress(true)
               if a > $FFFF
                  b := a & $FF
                  a := (a >> 8) & $FF
                  outa[a..b] := specialExpr
                  dira[a..b]~~
               else
                  outa[a] := specialExpr
                  dira[a]~~
            160: ' pause <expr> {,<expr>}
               if pauseTime == 0                 ' If no active pause time, set it
                  spaces                         '  with a minimum time of 50us
                  pauseTime := expr * 1000
                  if spaces == ","               ' First (or only) value is in ms
                     skipspaces
                     pauseTime += expr           ' Second value is in us
                  pauseTime #>= 50
               if pauseTime < 10_050             ' Normally pause at most 10ms at a time,
                  waitcnt(pauseTime * uS + cnt)  '  but, if that would leave < 50us,
                  pauseTime := 0                 '   pause the whole amount now
               else                             
                  a := pauseTime <# 10_000     
                  waitcnt(a * uS + cnt)          ' Otherwise, pause at most 10ms and
                  nextlineloc := thisLine        '  re-execute the PAUSE for the rest
                  pauseTime -= 10_000
            166: ' byte [ <expr> ] = <expr>
               a := getAddress(false)
               byte[a] := specialExpr
            167: ' word [ <expr> ] = <expr>
               a := getAddress(false)
               word[a] := specialExpr
            168: ' long [ <expr> ] = <expr>
               a := getAddress(false)
               long[a] := specialExpr
            170: ' phsa =
               PHSA := specialExpr
            171: ' phsb =
               PHSB := specialExpr
            172: ' frqa =
               FRQA := specialExpr
            173: ' frqb =
               FRQB := specialExpr
            174: ' ctra =
               CTRA := specialExpr
            175: ' ctrb =
               CTRB := specialExpr
            176: ' DISPLAY <expr> {,<expr>}
               spaces
               dsp.out(expr)
               repeat while spaces == ","
                  skipspaces
                  dsp.out(expr)
            178: ' LET <var> = <expr>
               nt := spaces
               if not isvar(nt)
                  abort @syn
               tp++
               vars[fixvar(nt)] := specialExpr
            179: ' STOP
               nextlineloc := eop-2
               return
            180: ' END
               nextlineloc := eop-2
               return
      else                                   
         abort(@syn)

pri bexec
   tokenize
   if spaces
      texec

pri repl | err
   clearall
   repeat
      err := \doline
      if err
         if curlineno => 0
            dsp.str(string("IN LINE "))
            dsp.dec(curlineno)
            dsp.out(" ")
         putlinet(err)
         nextlineloc := eop - 2         

pri doline | c
   curlineno := -1
   if key.peekkey == escapeKey
      key.getkey
      nextlineloc := eop-2
   if nextlineloc < eop-2
      curlineno := wordat(nextlineloc)
      tp := nextlineloc + 2
      nextlineloc := tp + strsize(tp) + 1
      texec
   else
      if fsrw.pclose < 0
         dsp.str(string("Error closing open file",dsp#Cr))
      pauseTime := 0
      dira := 0 ' Note: The display, keyboard, I2C, and SD FAT drivers all
      outa := 0 '       operate out of other cogs, so it's ok to clear these
      putlinet(string("OK"))
      getline
      c := spaces
      if "0" =< c and c =< "9"
         insertline
         nextlineloc := eop - 2
      else
         bexec      

PRI getAnyNumber | c, t
   case c := byte[tp]
      quote:
         if result := byte[++tp]
            if byte[++tp] == quote
              tp++
            else
               abort string("missing closing quote")
         else
            abort string("end of line in string")
      "$":
         c := byte[++tp]
         if (t := hexDigit(c)) < 0
            abort string("invalid hex character")
         result := t
         c := byte[++tp]
         repeat until (t := hexDigit(c)) < 0
            result := result << 4 | t
            c := byte[++tp]
      "%":
         c := byte[++tp]
         if not (c == "0" or c == "1")
            abort string("invalid binary character")
         result := c - "0"
         c := byte[++tp]
         repeat while c == "0" or c == "1"
            result := result << 1 | (c - "0")
            c := byte[++tp]
      "0".."9":
         result := c - "0"
         c := byte[++tp]
         repeat while c => "0" and c =< "9"
            result := result * 10 + c - "0"
            c := byte[++tp]
      other:
        abort string("invalid literal value")

PRI hexDigit(c)
'' Convert a hexadecimal character to the corresponding value or -1 if invalid.
   if c => "0" and c =< "9"
      return c - "0"
   if c => "A" and c =< "F"
      return c - "A" + 10
   if c => "a" and c =< "f"
      return c - "a" + 10
   return -1